from __future__ import annotations

import io
import struct
import typing

import numpy

from . import binary, type

MAX_LENGTH = (1 << 31) - 1


class BincodeSerializer(binary.Serializer):
    def __init__(self):
        super().__init__(output=io.BytesIO(), container_depth_budget=None)

    def serialize_f32(self, value: type.float32):
        self.output.write(struct.pack("<f", value))

    def serialize_f64(self, value: type.float64):
        self.output.write(struct.pack("<d", value))

    def serialize_len(self, value: int):
        if value > MAX_LENGTH:
            raise type.SerializationError("Length exceeds the maximum supported value.")
        self.output.write(int(value).to_bytes(8, "little", signed=False))

    def serialize_variant_index(self, value: int):
        self.output.write(int(value).to_bytes(4, "little", signed=False))

    def sort_map_entries(self, offsets: typing.List[int]):
        pass


class BincodeDeserializer(binary.Deserializer):
    def __init__(self, content):
        super().__init__(input=io.BytesIO(content), container_depth_budget=None)

    def deserialize_f32(self) -> type.float32:
        (value,) = struct.unpack("<f", self.read(4))
        return numpy.float32(value)

    def deserialize_f64(self) -> type.float64:
        (value,) = struct.unpack("<d", self.read(8))
        return numpy.float64(value)

    def deserialize_len(self) -> int:
        value = int.from_bytes(self.read(8), byteorder="little", signed=False)
        if value > MAX_LENGTH:
            raise type.DeserializationError(
                "Length exceeds the maximum supported value."
            )
        return value

    def deserialize_variant_index(self) -> int:
        return int.from_bytes(self.read(4), byteorder="little", signed=False)

    def check_that_key_slices_are_increasing(
        self,
        slice1: tuple[int, int],
        slice2: tuple[int, int],
    ):
        pass


def serialize(obj: typing.Any, obj_type) -> bytes:
    serializer = BincodeSerializer()
    serializer.serialize_any(obj, obj_type)
    return serializer.get_buffer()


def deserialize(content: bytes, obj_type) -> tuple[typing.Any, bytes]:
    deserializer = BincodeDeserializer(content)
    value = deserializer.deserialize_any(obj_type)
    return value, deserializer.get_remaining_buffer()



from __future__ import annotations

import dataclasses
import typing


@dataclasses.dataclass
class RingStatus:
    system_time: float
    read_range: tuple[int, int]
    write_range: tuple[int, int]
    ring_length: int
    # current_t is None if the mode is raw
    current_t: typing.Optional[int]

    def backlog(self) -> int:
        return (
            self.write_range[0] + self.ring_length - self.read_range[1]
        ) % self.ring_length

    def raw_packets(self) -> int:
        return (
            self.read_range[1] + self.ring_length - self.read_range[0]
        ) % self.ring_length


@dataclasses.dataclass
class Status:
    system_time: float
    # ring is None if no data became available before iterator_timeout
    # ring may only be None if iterator_timeout is not None
    ring: typing.Optional[RingStatus]

    def delay(self) -> typing.Optional[float]:
        if self.ring is None:
            return None
        return self.system_time - self.ring.system_time


@dataclasses.dataclass
class StatusNonOptional:
    system_time: float
    ring: RingStatus

    def delay(self) -> float:
        ...



from __future__ import annotations

import dataclasses
import enum
import types
import typing

import numpy

from ... import serde
from ... import status
from .. import enums


@dataclasses.dataclass
class Biases:
    pr: serde.type.uint8 = 0x7C
    fo: serde.type.uint8 = 0x53
    hpf: serde.type.uint8 = 0x00
    diff_on: serde.type.uint8 = 0x66
    diff: serde.type.uint8 = 0x4D
    diff_off: serde.type.uint8 = 0x49
    inv: serde.type.uint8 = 0x5B
    refr: serde.type.uint8 = 0x14
    reqpuy: serde.type.uint8 = 0x8C
    reqpux: serde.type.uint8 = 0x7C
    sendreqpdy: serde.type.uint8 = 0x94
    unknown_1: serde.type.uint8 = 0x74
    unknown_2: serde.type.uint8 = 0x51

    def serialize(self) -> bytes:
        return serde.bincode.serialize(self, Biases)


class Clock(enum.Enum):
    INTERNAL = 0
    INTERNAL_WITH_OUTPUT_ENABLED = 1
    EXTERNAL = 2

    def serialize(self) -> bytes:
        return serde.bincode.serialize(self, Clock)


@dataclasses.dataclass
class RateLimiter:
    reference_period_us: serde.type.uint16
    maximum_events_per_period: serde.type.uint32

    def serialize(self) -> bytes:
        return serde.bincode.serialize(self, RateLimiter)


@dataclasses.dataclass
class Configuration:
    biases: Biases = dataclasses.field(default_factory=Biases)
    x_mask: tuple[
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
    ] = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    y_mask: tuple[
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
    ] = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    pixel_mask: tuple[
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
    ] = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    mask_intersection_only: bool = False
    enable_external_trigger: bool = True
    clock: Clock = Clock.INTERNAL
    rate_limiter: typing.Optional[RateLimiter] = None
    enable_output: bool = True

    def serialize(self) -> bytes:
        return serde.bincode.serialize(self, Configuration)

    @staticmethod
    def type() -> str:
        return "prophesee_evk4"


@dataclasses.dataclass
class UsbConfiguration:
    buffer_size: serde.type.uint64 = 131072
    ring_size: serde.type.uint64 = 4096
    transfer_queue_size: serde.type.uint64 = 32
    allow_dma: bool = False

    def serialize(self) -> bytes:
        return serde.bincode.serialize(self, UsbConfiguration)


@dataclasses.dataclass(frozen=True)
class Properties:
    width: serde.type.uint16 = 1280
    height: serde.type.uint16 = 720


class Device(typing.Protocol):
    def __enter__(self) -> "Device":
        ...

    def __exit__(
        self,
        exception_type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool:
        ...

    def __iter__(self) -> "Device":
        ...

    def __next__(self) -> tuple[status.StatusNonOptional, dict[str, numpy.ndarray[typing.Any, numpy.dtype[numpy.void]]]]:
        ...

    def clear_backlog(self, until: int):
        ...

    def name(self) -> typing.Literal[enums.Name.PROPHESEE_EVK4]:
        ...

    def properties(self) -> Properties:
        ...

    def serial(self) -> str:
        ...

    def chip_firmware_configuration(self) -> Configuration:
        ...

    def speed(self) -> enums.Speed:
        ...

    def update_configuration(self, configuration: Configuration):
        ...


class DeviceOptional(typing.Protocol):
    def __enter__(self) -> "DeviceOptional":
        ...

    def __exit__(
        self,
        exception_type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool:
        ...

    def __iter__(self) -> "DeviceOptional":
        ...

    def __next__(self) -> tuple[status.Status, typing.Optional[dict[str, numpy.ndarray[typing.Any, numpy.dtype[numpy.void]]]]]:
        ...

    def clear_backlog(self, until: int):
        ...

    def name(self) -> typing.Literal[enums.Name.PROPHESEE_EVK4]:
        ...

    def properties(self) -> Properties:
        ...

    def serial(self) -> str:
        ...

    def chip_firmware_configuration(self) -> Configuration:
        ...

    def speed(self) -> enums.Speed:
        ...

    def update_configuration(self, configuration: Configuration):
        ...


class DeviceRaw(typing.Protocol):
    def __enter__(self) -> "DeviceRaw":
        ...

    def __exit__(
        self,
        exception_type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool:
        ...

    def __iter__(self) -> "DeviceRaw":
        ...

    def __next__(self) -> tuple[status.StatusNonOptional, bytes]:
        ...

    def clear_backlog(self, until: int):
        ...

    def name(self) -> typing.Literal[enums.Name.PROPHESEE_EVK4]:
        ...

    def properties(self) -> Properties:
        ...

    def serial(self) -> str:
        ...

    def chip_firmware_configuration(self) -> Configuration:
        ...

    def speed(self) -> enums.Speed:
        ...

    def update_configuration(self, configuration: Configuration):
        ...


class DeviceRawOptional(typing.Protocol):
    def __enter__(self) -> "DeviceRawOptional":
        ...

    def __exit__(
        self,
        exception_type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool:
        ...

    def __iter__(self) -> "DeviceRawOptional":
        ...

    def __next__(self) -> tuple[status.Status, typing.Optional[bytes]]:
        ...

    def clear_backlog(self, until: int):
        ...

    def name(self) -> typing.Literal[enums.Name.PROPHESEE_EVK4]:
        ...

    def properties(self) -> Properties:
        ...

    def serial(self) -> str:
        ...

    def chip_firmware_configuration(self) -> Configuration:
        ...

    def speed(self) -> enums.Speed:
        ...

    def update_configuration(self, configuration: Configuration):
        ...



import pathlib

import neuromorphic_drivers as nd
import numpy as np
import plotly.express

dirname = pathlib.Path(__file__).resolve().parent

pixel_mask = nd.PropheseeEvk4PixelMask()
pixel_mask.set(63, 1026, 664)
nd.write_mask_as_png(
    pixel_mask.pixels(),
    dirname / "evk4_mask.png",
)

print(pixel_mask.pixel_mask())

counts = np.zeros(
    (nd.prophesee_evk4.Properties().width, nd.prophesee_evk4.Properties().height),
    dtype=np.uint64,
)

with nd.open(
    configuration=nd.prophesee_evk4.Configuration(
        biases=nd.prophesee_evk4.Biases(
            diff_off=50,
        ),
        pixel_mask=pixel_mask.pixel_mask(),
    )
) as device:
    for status, packet in device:
        if "dvs_events" in packet:
            np.add.at(counts, (packet["dvs_events"]["x"], packet["dvs_events"]["y"]), 1)
            if packet["dvs_events"]["t"][-1] > 2000000:
                break

log_counts = np.flipud(np.log((counts + 1).astype(np.float64)).transpose())
figure = plotly.express.imshow(log_counts, color_continuous_scale="thermal")
figure.update_layout(
    width=nd.prophesee_evk4.Properties().width,
    height=nd.prophesee_evk4.Properties().height,
    margin=dict(l=0, r=0, b=0, t=0),
    xaxis={"visible": False, "showticklabels": False},
    yaxis={"visible": False, "showticklabels": False},
    showlegend=False,
)
figure.update_coloraxes(showscale=False)
figure.write_image(dirname / f"evk4_mask_accumulated.png")



import neuromorphic_drivers as nd
import numpy as np

nd.print_device_list()

devices: list[nd.GenericDevice] = []
for listed_device in nd.list_devices():
    device = nd.open(serial=listed_device.serial)
    devices.append(device)
    print(listed_device.serial, devices[0].serial(), devices[0].properties())

backlogs = np.array([0 for _ in devices])
while True:
    index = np.argmax(backlogs)
    status, packet = devices[index].__next__()
    backlog = status.ring.backlog()
    print(f"{index}: {round(status.delay() * 1e6)} µs, backlog: {backlog}")
    backlogs[:] += 1
    backlogs[index] = backlog



from __future__ import annotations

import pathlib
import struct
import typing
import zlib

import numpy

from . import serde


def write_mask_as_png(
    pixels: numpy.ndarray, path: typing.Union[str, bytes, pathlib.Path]
):
    def pack(tag: bytes, data: bytes):
        content = tag + data
        return b"".join(
            (
                struct.pack(">I", len(data)),
                content,
                struct.pack(">I", 0xFFFFFFFF & zlib.crc32(content)),
            )
        )

    with open(path, "wb") as file:
        file.write(
            b"".join(
                [
                    b"\x89PNG\r\n\x1a\n",
                    pack(
                        b"IHDR",
                        struct.pack(
                            ">IIBBBBB",
                            pixels.shape[1],
                            pixels.shape[0],
                            1,
                            0,
                            0,
                            0,
                            0,
                        ),
                    ),
                    pack(
                        b"IDAT",
                        zlib.compress(
                            numpy.hstack(
                                (
                                    numpy.zeros(
                                        (pixels.shape[0], 1), dtype=numpy.uint8
                                    ),
                                    numpy.packbits(pixels).reshape(
                                        (pixels.shape[0], pixels.shape[1] // 8)
                                    ),
                                )
                            ).tobytes(),
                            9,
                        ),
                    ),
                    pack(b"IEND", b""),
                ]
            )
        )


class RowColumnMask:
    def __init__(self, width: serde.type.uint16, height: serde.type.uint16, set: bool):
        self.x_unpacked = numpy.full(width, fill_value=set, dtype=bool)
        self.y_unpacked = numpy.full(height, fill_value=set, dtype=bool)

    def clear_x_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.x_unpacked[index] = False

    def set_x_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.x_unpacked[index] = True

    def clear_y_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.y_unpacked[index] = False

    def set_y_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.x_unpacked[index] = True

    def clear_rectangle(
        self,
        x: serde.type.uint16,
        y: serde.type.uint16,
        width: serde.type.uint16,
        height: serde.type.uint16,
    ):
        self.clear_x_range(start=x, stop=x + width, step=1)
        self.clear_y_range(start=y, stop=y + height, step=1)

    def x_mask(self) -> tuple[serde.type.uint64, ...]:
        uint8 = numpy.packbits(self.x_unpacked, bitorder="little")
        if len(uint8) % 8 != 0:
            uint8 = numpy.append(
                uint8, numpy.zeros(8 - len(uint8) % 8, dtype=numpy.uint8)
            )
        return tuple(uint8.view(dtype="=u8").tolist())

    def y_mask(self) -> tuple[serde.type.uint64, ...]:
        uint8 = numpy.packbits(self.y_unpacked, bitorder="little")
        if len(uint8) % 8 != 0:
            uint8 = numpy.append(
                uint8, numpy.zeros(8 - len(uint8) % 8, dtype=numpy.uint8)
            )
        return tuple(uint8.view(dtype="=u8").tolist())

    def pixels(self) -> numpy.ndarray[typing.Any, numpy.dtype[numpy.bool_]]:
        result = numpy.full(
            (len(self.y_unpacked), len(self.x_unpacked)), fill_value=False, dtype=bool
        )
        result[:, self.x_unpacked] = True
        result[numpy.flip(self.y_unpacked), :] = True
        return numpy.bitwise_not(result)


class PropheseeEvk4PixelMask:
    def __init__(self):
        self.pixel_masks = numpy.zeros(
            64,
            dtype=[("x", numpy.uint16), ("y", numpy.uint16), ("set", numpy.bool_)],
        )

    def set(self, index: int, x: int, y: int):
        assert index >= 0 and index < 64
        assert x >= 0 and x < 1280
        assert y >= 0 and y < 720
        self.pixel_masks["x"][index] = x
        self.pixel_masks["y"][index] = y
        self.pixel_masks["set"][index] = True

    def clear(self, index: int):
        assert index < 64
        self.pixel_masks["x"][index] = 0
        self.pixel_masks["y"][index] = 0
        self.pixel_masks["set"][index] = False

    def pixels(self) -> numpy.ndarray[typing.Any, numpy.dtype[numpy.bool_]]:
        result = numpy.full(
            (720, 1280),
            fill_value=False,
            dtype=bool,
        )
        selected_pixel_masks = self.pixel_masks[self.pixel_masks["set"]]
        result[720 - 1 - selected_pixel_masks["y"], selected_pixel_masks["x"]] = True
        return result

    def pixel_mask(
        self,
    ) -> tuple[
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
        serde.type.uint64,
    ]:
        codes = (
            self.pixel_masks["x"].astype(numpy.uint64)
            + self.pixel_masks["y"].astype(numpy.uint64) * 1280
            + 1
        )
        codes[numpy.logical_not(self.pixel_masks["set"])] = 0
        result = numpy.zeros(21, dtype=numpy.uint64)
        for index in range(0, 63):
            result[index // 3] |= numpy.left_shift(
                codes[index], numpy.uint64((index % 3) * 21)
            )
        for bit in range(0, 21):
            result[bit] |= numpy.left_shift(
                numpy.bitwise_and(
                    numpy.right_shift(codes[63], numpy.uint64(bit)),
                    numpy.uint64(1),
                ),
                numpy.uint64(63),
            )
        return tuple(result.tolist())  # type: ignore
